"""ManagedAccount Module, all the logic to retrieve managed accounts from PS API"""

import logging

import requests
from cerberus import Validator

from secrets_safe_library import exceptions, utils
from secrets_safe_library.core import APIObject


class ManagedAccount(APIObject):

    _authentication = None
    _logger = None
    _separator = None
    _sign_app_out_error_message = "Error in sign_app_out"

    def __init__(self, authentication, logger=None, separator="/"):
        self._authentication = authentication
        self._logger = logger

        # Schema rules used for validations
        self._schema = {
            "system_name": {"type": "string", "maxlength": 129, "nullable": True},
            "account_name": {"type": "string", "maxlength": 246, "nullable": True},
            "system_id": {"type": "integer", "nullable": True},
            "workgroup_name": {"type": "string", "maxlength": 256, "nullable": True},
            "application_display_name": {
                "type": "string",
                "maxlength": 256,
                "nullable": True,
            },
            "ip_address": {"type": "string", "maxlength": 45, "nullable": True},
            "type": {
                "type": "string",
                "allowed": [
                    "system",
                    "recent",
                    "domainlinked",
                    "database",
                    "cloud",
                    "application",
                ],
                "nullable": True,
            },
            "limit": {"type": "integer", "min": 1, "max": 1000, "default": 1000},
            "offset": {"type": "integer", "min": 0, "default": 0},
        }
        self._validator = Validator(self._schema)

        if len(separator.strip()) != 1:
            raise exceptions.LookupError(f"Invalid separator: {separator}")
        self._separator = separator

    def get_secret(self, path):
        """
        Get Managed account by path
        Arguments:
            path
        Returns:
            Retrieved managed account string
        """

        utils.print_log(
            self._logger,
            "Running get_secret method in ManagedAccount class",
            logging.DEBUG,
        )
        managed_account_dict = self.managed_account_flow([path])
        return managed_account_dict[path]

    def get_secret_with_metadata(self, path):
        """
        Get Managed account with metadata by path
        Arguments:
            path
        Returns:
             Retrieved managed account in dict format
        """

        utils.print_log(
            self._logger,
            "Running get_secret method in ManagedAccount class",
            logging.DEBUG,
        )
        managed_account_dict = self.managed_account_flow([path], get_metadata=True)
        return managed_account_dict

    def get_secrets(self, paths):
        """
        Get Managed accounts by paths
        Arguments:
            paths list
        Returns:
            Retrieved managed account in dict format
        """

        utils.print_log(
            self._logger,
            "Running get_secrets method in ManagedAccount class",
            logging.INFO,
        )
        managed_account_dict = self.managed_account_flow(paths)
        return managed_account_dict

    def get_secrets_with_metadata(self, paths):
        """
        Get Managed accounts with metadata by paths
        Arguments:
            paths list
        Returns:
            Retrieved managed account in dict format
        """

        utils.print_log(
            self._logger,
            "Running get_secrets method in ManagedAccount class",
            logging.INFO,
        )
        managed_account_dict = self.managed_account_flow(paths, get_metadata=True)
        return managed_account_dict

    def get_request_id(self, system_id, account_id):
        create_request_response = self.create_request(system_id, account_id)
        request_id = create_request_response.json()
        return request_id

    def managed_account_flow(self, paths, get_metadata=False):
        """
        Managed account by path flow
        Arguments:
            paths list
        Returns:
            Response (Dict)
        """

        response = {}

        for path in paths:

            utils.print_log(
                self._logger,
                f"**************** managed account path: {path} ****************",
                logging.INFO,
            )
            data = path.split(self._separator)

            if len(data) != 2:
                raise exceptions.LookupError(
                    f"Invalid managed account path: {path}. Use '{self._separator}' as "
                    f"a delimiter: system_name{self._separator}managed_account_name"
                )

            system_name = data[0]
            managed_account_name = data[1]

            manage_account = self.get_managed_accounts(
                system_name=system_name, account_name=managed_account_name
            )

            if get_metadata:
                response[f"{path}-metadata"] = manage_account

            utils.print_log(
                self._logger, "Managed account info retrieved", logging.DEBUG
            )

            request_id = self.get_request_id(
                manage_account["SystemId"], manage_account["AccountId"]
            )

            utils.print_log(
                self._logger,
                f"Request id retrieved: {'*' * len(str(request_id))}",
                logging.DEBUG,
            )

            if not request_id:
                raise exceptions.LookupError("Request Id not found")

            credential = self.get_credential_by_request_id(request_id)

            response[path] = credential

            utils.print_log(
                self._logger, "Credential was successfully retrieved", logging.INFO
            )

            self.request_check_in(request_id)
        return response

    def create_request(self, system_id: int, account_id: int) -> requests.Response:
        """
        Create request by system ID and account ID.

        API: POST Requests

        Args:
            system_id (int): ID of the managed system to request.
            account_id (int): ID of the managed account to request.

        Returns:
            request.Response: Response object.
        """
        payload = {
            "SystemID": system_id,
            "AccountID": account_id,
            "DurationMinutes": 5,
            "Reason": "Secrets Safe Integration",
            "ConflictOption": "reuse",
        }

        endpoint = "/Requests"
        utils.print_log(
            self._logger, f"Calling create_request endpoint: {endpoint}", logging.DEBUG
        )
        response = self._run_post_request(
            endpoint,
            payload=payload,
            expected_status_code=[200, 201],
            include_api_version=False,
        )
        return response

    def get_credential_by_request_id(self, request_id: int):
        """
        Retrieves the credentials for an approved and active (not expired) credentials
        release request.

        API: GET Credentials/{requestId}

        Args:
            request_id (int): The request ID.

        Returns:
            requests.Response: The response object containing the credential details.
        """

        endpoint = f"/Credentials/{request_id}"
        print_url = (
            f"{self._authentication._api_url}/Credentials/{'*' * len(str(request_id))}"
        )

        utils.print_log(
            self._logger,
            f"Calling get_credential_by_request_id endpoint: {print_url}",
            logging.DEBUG,
        )
        response = self._run_get_request(endpoint, include_api_version=False)

        credential = response.json()
        return credential

    def request_check_in(self, request_id: int, reason: str = "") -> None:
        """
        Checks-in/releases a request before it has expired.

        API: PUT Requests/{id}/checkin.

        Args:
            request_id (int): The ID of the request to check-in/release.
            reason (str, optional): A reason or comment why the request is being
                approved. Max string length is 1000.

        Returns:
            None
        """
        endpoint = f"/Requests/{request_id}/checkin"
        print_url = (
            f"{self._authentication._api_url}/Requests/"
            f"{'*' * len(str(request_id))}/checkin"
        )

        utils.print_log(
            self._logger,
            f"Calling request_check_in endpoint: {print_url}",
            logging.DEBUG,
        )
        _ = self._run_put_request(
            endpoint,
            payload={},
            expected_status_code=204,
        )

        utils.print_log(self._logger, "Request checked in", logging.DEBUG)

    def list_by_managed_system(self, managed_system_id: int) -> list:
        """
        Returns a list of managed accounts by managed system ID.

        API: GET ManagedSystems/{systemID}/ManagedAccounts

        Args:
            managed_system_id (int): Managed system ID.

        Returns:
            list: List of managed accounts.
        """

        endpoint = f"/managedsystems/{managed_system_id}/managedaccounts"

        utils.print_log(
            self._logger,
            f"Calling list_by_managed_system endpoint: {endpoint}",
            logging.DEBUG,
        )
        response = self._run_get_request(endpoint)

        return response.json()

    def get_by_id(self, managed_account_id: int) -> dict:
        """
        Returns a managed account by ID.

        API: GET ManagedAccounts/{id}

        Args:
            managed_account_id (int): Managed account ID.

        Returns:
            dict: Managed account details.
        """

        endpoint = f"/managedaccounts/{managed_account_id}"

        utils.print_log(
            self._logger, f"Calling get_by_id endpoint: {endpoint}", logging.DEBUG
        )
        response = self._run_get_request(endpoint)

        return response.json()

    def get_managed_accounts(
        self,
        *,
        system_name: str = None,
        account_name: str = None,
        system_id: str = None,
        workgroup_name: str = None,
        application_display_name: str = None,
        ip_address: str = None,
        type: str = None,
        limit: int = 1000,
        offset: int = 0,
    ) -> list | dict:
        """
        Get managed account(s) by system name, account name and other fields.

        API: GET ManagedAccounts.

        Args:
            system_name (str): The name of the system where the account is managed.
            account_name (str): The name of the account to retrieve.
            system_id (str): The ID of the managed system.
            workgroup_name (str): The name of the workgroup.
            application_display_name (str): The display name of the application.
            ip_address (str): The IP address of the managed asset.
            type (str): The type of the managed account to return. Options are:
                system, recent, domainlinked, database, cloud, application.
            limit (int): The number of records to return. Default is 1000.
            offset (int): The number of records to skip before returning records.
                Default is 0.

        Returns:
            list | dict: List of managed accounts if multiple accounts are found,
                otherwise, a dictionary containing a single accound data will be
                returned.

        Raises:
            exceptions.OptionsError: If provided arguments are not valid.
        """
        attributes = {
            "system_name": system_name,
            "account_name": account_name,
            "system_id": system_id,
            "workgroup_name": workgroup_name,
            "application_display_name": application_display_name,
            "ip_address": ip_address,
            "type": type,
            "limit": limit,
            "offset": offset,
        }

        params = {key.replace("_", ""): value for key, value in attributes.items()}

        if self._validator.validate(attributes, update=True):
            query_string = self.make_query_string(params)
            endpoint = f"/managedaccounts?{query_string}"
            utils.print_log(
                self._logger,
                f"Calling get_managed_accounts endpoint {endpoint}",
                logging.DEBUG,
            )
            response = self._run_get_request(endpoint, include_api_version=False)
            return response.json()
        else:
            raise exceptions.OptionsError(f"Please check: {self._validator.errors}")

    def list_by_smart_rule_id(self, smart_rule_id: int) -> list:
        """
        Returns a list of managed accounts by Smart Rule ID.

        API: GET SmartRules/{smartRuleID}/ManagedAccounts

        Args:
            smart_rule_id (int): Smart Rule ID.

        Returns:
            list: List of managed accounts.
        """

        endpoint = f"/smartrules/{smart_rule_id}/managedaccounts"

        utils.print_log(
            self._logger,
            f"Calling list_by_smart_rule_id endpoint: {endpoint}",
            logging.DEBUG,
        )
        response = self._run_get_request(endpoint)

        return response.json()

    def list_by_quick_rule_id(self, quick_rule_id: int) -> list:
        """
        Returns a list of managed accounts by Quick Rule ID.

        API: GET QuickRules/{quickRuleID}/ManagedAccounts

        Args:
            quick_rule_id (int): Quick rule ID.

        Returns:
            list: List of managed accounts.
        """

        endpoint = f"/quickrules/{quick_rule_id}/managedaccounts"

        utils.print_log(
            self._logger,
            f"Calling list_by_quick_rule_id endpoint: {endpoint}",
            logging.DEBUG,
        )
        response = self._run_get_request(endpoint)

        return response.json()



"""SmartRules module, all the logic to manage Smart Rules from BeyondInsight API"""

import logging

from cerberus import Validator

from secrets_safe_library import exceptions, utils
from secrets_safe_library.authentication import Authentication
from secrets_safe_library.core import APIObject


class SmartRule(APIObject):

    def __init__(self, authentication: Authentication, logger: logging.Logger = None):
        super().__init__(authentication, logger)

        # Schema rules used for validations
        self._schema = {
            "title": {"type": "string", "maxlength": 256, "nullable": True},
            "rule_id": {"type": "integer", "nullable": True},
        }
        self._validator = Validator(self._schema)

    def get_smart_rule_by_id(self, rule_id: int) -> dict:
        """
        Returns a SmartRule by ID.

        API: GET SmartRules/{id}

        Args:
            rule_id (int): The SmartRule ID.

        Returns:
            dict: SmartRule.
        """

        attributes = {"rule_id": rule_id}

        if not self._validator.validate(attributes, update=True):
            raise exceptions.OptionsError(f"Please check: {self._validator.errors}")

        endpoint = f"/smartrules/{rule_id}"

        utils.print_log(
            self._logger,
            f"Calling get_smart_rule_by_id endpoint: {endpoint}",
            logging.DEBUG,
        )
        response = self._run_get_request(endpoint, include_api_version=False)

        return response.json()

    def get_smart_rule_by_title(self, title: str) -> dict:
        """
        Returns a SmartRule by Title.

        API: GET SmartRules?title={title}

        Args:
            title (str): The SmartRule title.

        Returns:
            dict: SmartRule.
        """

        attributes = {"title": title}

        if not self._validator.validate(attributes, update=True):
            raise exceptions.OptionsError(f"Please check: {self._validator.errors}")

        query_string = self.make_query_string(attributes)
        endpoint = f"/smartrules/?{query_string}"

        utils.print_log(
            self._logger,
            f"Calling get_smart_rule_by_title endpoint: {endpoint}",
            logging.DEBUG,
        )
        response = self._run_get_request(endpoint, include_api_version=False)

        return response.json()

    def list_assets_by_smart_rule_id(
        self, smart_rule_id: int, limit: int = None, offset: int = None
    ) -> list:
        """
        Returns a list of assets for the given smart rule ID.

        API: GET SmartRules/{id}/Assets

        Args:
            smart_rule_id (int): The Smart Rule ID.
            limit (int, optional): limit the results.
            offset (int, optional): skip the first (offset) number of assets.

        Returns:
            list: List of assets for the specified smart_rule_id.
        """

        attributes = {"rule_id": smart_rule_id}

        if not self._validator.validate(attributes, update=True):
            raise exceptions.OptionsError(f"Please check: {self._validator.errors}")

        params = {
            "limit": limit,
            "offset": offset,
        }
        query_string = self.make_query_string(params)

        endpoint = f"/smartrules/{smart_rule_id}/assets?{query_string}"

        utils.print_log(
            self._logger,
            f"Calling list_assets_by_smart_rule_id endpoint: {endpoint}",
            logging.DEBUG,
        )
        response = self._run_get_request(endpoint)

        return response.json()

    def get_quick_rule_by_id(self, rule_id: int) -> dict:
        """
        Returns a QuickRule by ID.

        API: GET QuickRules/{id}

        Args:
            rule_id (int): The QuickRule ID.

        Returns:
            dict: QuickRule.
        """

        attributes = {"rule_id": rule_id}

        if not self._validator.validate(attributes, update=True):
            raise exceptions.OptionsError(f"Please check: {self._validator.errors}")

        endpoint = f"/quickrules/{rule_id}"

        utils.print_log(
            self._logger,
            f"Calling get_quick_rule_by_id endpoint: {endpoint}",
            logging.DEBUG,
        )
        response = self._run_get_request(endpoint, include_api_version=False)

        return response.json()

    def get_quick_rule_by_title(self, title: str) -> dict:
        """
        Returns a QuickRule by Title.

        API: GET QuickRules?title={title}

        Args:
            title (str): The QuickRule title.

        Returns:
            dict: QuickRule.
        """

        attributes = {"title": title}

        if not self._validator.validate(attributes, update=True):
            raise exceptions.OptionsError(f"Please check: {self._validator.errors}")

        query_string = self.make_query_string(attributes)
        endpoint = f"/quickrules/?{query_string}"

        utils.print_log(
            self._logger,
            f"Calling get_quick_rule_by_title endpoint: {endpoint}",
            logging.DEBUG,
        )
        response = self._run_get_request(endpoint, include_api_version=False)

        return response.json()



"""
This file contains the fields mapping for each endpoint and version related to Safes.

More info for the fields included in each endpoint and version:
- https://docs.beyondtrust.com/bips/docs/passwordsafe-api
- https://docs.beyondtrust.com/bips/docs/secrets-safe-api
"""

from secrets_safe_library.constants.endpoints import (
    GET_ASSETS_ID,
    GET_ASSETS_ID_ATTRIBUTES,
    GET_WORKGROUPS_ID_ASSETS,
    GET_WORKGROUPS_NAME_ASSETS_NAME,
)
from secrets_safe_library.constants.versions import Version

fields = {
    GET_WORKGROUPS_ID_ASSETS: {
        Version.DEFAULT: [
            "WorkgroupID",
            "AssetID",
            "AssetName",
            "DnsName",
            "DomainName",
            "IPAddress",
            "MacAddress",
            "AssetType",
            "OperatingSystem",
            "CreateDate",
            "LastUpdateDate",
        ],
    },
    GET_ASSETS_ID: {
        Version.DEFAULT: [
            "WorkgroupID",
            "AssetID",
            "AssetName",
            "DnsName",
            "DomainName",
            "IPAddress",
            "MacAddress",
            "AssetType",
            "OperatingSystem",
            "CreateDate",
            "LastUpdateDate",
        ],
    },
    GET_WORKGROUPS_NAME_ASSETS_NAME: {
        Version.DEFAULT: [
            "WorkgroupID",
            "AssetID",
            "AssetName",
            "DnsName",
            "DomainName",
            "IPAddress",
            "MacAddress",
            "AssetType",
            "OperatingSystem",
            "CreateDate",
            "LastUpdateDate",
        ],
    },
    GET_ASSETS_ID_ATTRIBUTES: {
        Version.DEFAULT: [
            "AttributeID",
            "AttributeTypeID",
            "ParentAttributeID",
            "ShortName",
            "LongName",
            "Description",
            "ValueInt",
            "IsReadOnly",
        ],
    },
}



"""
This file contains the fields mapping for each endpoint and version related to
WORKGROUPS.

More info for the fields included in each endpoint and version:
- https://docs.beyondtrust.com/bips/docs/passwordsafe-api
- https://docs.beyondtrust.com/bips/docs/secrets-safe-api
"""

from secrets_safe_library.constants.endpoints import (
    GET_WORKGROUPS,
    GET_WORKGROUPS_ID,
    GET_WORKGROUPS_NAME,
    POST_WORKGROUPS,
)
from secrets_safe_library.constants.versions import Version

fields = {
    GET_WORKGROUPS: {
        Version.DEFAULT: [
            "OrganizationID",
            "Name",
        ],
    },
    GET_WORKGROUPS_ID: {
        Version.DEFAULT: [
            "OrganizationID",
            "ID",
            "Name",
        ],
    },
    GET_WORKGROUPS_NAME: {
        Version.DEFAULT: [
            "OrganizationID",
            "ID",
            "Name",
        ],
    },
    POST_WORKGROUPS: {
        Version.DEFAULT: {
            "OrganizationID": str,
            "Name": str,
        },
    },
}



"""
This file contains the fields mapping for each endpoint and version related to Smart
Rules.

More info for the fields included in each endpoint and version:
- https://docs.beyondtrust.com/bips/docs/beyondinsight-api
- https://docs.beyondtrust.com/bips/docs/passwordsafe-api
- https://docs.beyondtrust.com/bips/docs/secrets-safe-api
"""

from secrets_safe_library.constants.endpoints import GET_SMARTRULES_ID
from secrets_safe_library.constants.versions import Version

fields = {
    GET_SMARTRULES_ID: {
        Version.DEFAULT: [
            "SmartRuleID",
            "OrganizationID ",
            "Title",
            "Description",
            "Category",
            "Status",
            "LastProcessedDate",
            "IsReadOnly",
            "RuleType",
        ],
    },
}



import json
import logging
import os
from abc import ABC, abstractmethod
from typing import List
from urllib.parse import parse_qs, urlencode, urlparse, urlunparse

import requests

from secrets_safe_library import exceptions, utils
from secrets_safe_library.authentication import Authentication


class APIObjectInterface(ABC):
    @abstractmethod
    def _run_get_request(
        self, endpoint: str, include_api_version: bool, expected_status_code: int
    ) -> requests.Response:
        pass

    @abstractmethod
    def _run_post_request(
        self,
        endpoint: str,
        payload: dict,
        include_api_version: bool,
        expected_status_code: int,
    ) -> requests.Response:
        pass

    @abstractmethod
    def _run_post_file_request(
        self,
        endpoint: str,
        payload: dict,
        include_api_version: bool,
        expected_status_code: int,
        file_path: str = None,
    ) -> requests.Response:
        pass

    @abstractmethod
    def _run_put_request(
        self,
        endpoint: str,
        payload: dict,
        include_api_version: bool,
        expected_status_code: int,
    ) -> requests.Response:
        pass

    @abstractmethod
    def _run_put_file_request(
        self,
        endpoint: str,
        payload: dict,
        include_api_version: bool,
        expected_status_code: int,
        file_path: str = None,
    ) -> requests.Response:
        pass

    @abstractmethod
    def _run_delete_request(
        self, endpoint: str, expected_status_code: int
    ) -> requests.Response:
        pass

    @abstractmethod
    def make_query_string(self, params: dict) -> str:
        pass


class APIObject(APIObjectInterface):
    _authentication: Authentication
    _logger: logging.Logger

    def __init__(self, authentication: Authentication, logger: logging.Logger = None):
        self._authentication = authentication
        self._logger = logger

    def __is_successful_response(
        self, status_code: int, expected_status_code: int | List[int]
    ) -> bool:
        """
        Checks if the given status code matches the expected status code(s).

        Args:
            status_code (int): The actual status code to validate.
            expected_status_code (int | List[int]): The expected status code or a list
                of acceptable status codes.

        Returns:
            bool: True if the status code matches the expected value(s), False
            otherwise.
        """
        return (
            status_code == expected_status_code
            if isinstance(expected_status_code, int)
            else status_code in expected_status_code
        )

    def _run_get_request(
        self,
        endpoint: str,
        include_api_version: bool = True,
        expected_status_code: int | List[int] = 200,
    ) -> requests.Response:
        """
        Shortcut function to make a GET request to the specified endpoint using
        Authentication object.

        Args:
            endpoint (str): The endpoint to call using GET HTTP verb.
            include_api_version (bool): Whether to include API version in request URL.
            expected_status_code (int | List[int], optional): Expected status code to
                consider the request was successful.

        Returns:
            request.Response: Response object.

        Raises:
            exceptions.LookupError: When get request fails.
        """
        url = self._create_url(endpoint, include_api_version)

        utils.print_log(self._logger, f"GET request to URL: {url}", logging.DEBUG)

        response = self._authentication._req.get(
            url,
            timeout=(
                self._authentication._timeout_connection_seconds,
                self._authentication._timeout_request_seconds,
            ),
        )

        if not self.__is_successful_response(
            response.status_code, expected_status_code
        ):
            if not self._authentication.sign_app_out():
                utils.print_log(
                    self._logger, "Error running get request", logging.ERROR
                )
            raise exceptions.LookupError(
                f"Error running get request, message: {response.text}, statuscode: "
                f"{response.status_code}"
            )

        return response

    def _run_post_request(
        self,
        endpoint: str,
        payload: dict,
        include_api_version: bool = True,
        expected_status_code: int | List[int] = 201,
    ) -> requests.Response:
        """
        Shortcut function to make a POST request to the specified endpoint using
        Authentication object.

        Args:
            endpoint (str): The endpoint to call using POST HTTP verb.
            payload (dict): Payload to send with the POST request.
            include_api_version (bool): Whether to include API version in request URL.
            expected_status_code (int | List[int], optional): Expected status code to
                consider the request was successful.

        Returns:
            request.Response: Response object.

        Raises:
            exceptions.CreationError: Raised when creation failed, because response
                status code is different to expected_status_code.
        """
        url = self._create_url(endpoint, include_api_version)

        utils.print_log(self._logger, f"Calling URL: {url}", logging.DEBUG)

        response = self._authentication._req.post(
            url=url,
            json=payload,
            timeout=(
                self._authentication._timeout_connection_seconds,
                self._authentication._timeout_request_seconds,
            ),
        )

        if not self.__is_successful_response(
            response.status_code, expected_status_code
        ):
            if not self._authentication.sign_app_out():
                utils.print_log(self._logger, "Error in post request", logging.ERROR)
            raise exceptions.CreationError(
                f"Error running post request, message: {response.text}, statuscode: "
                f"{response.status_code}"
            )

        return response

    def _run_post_file_request(
        self,
        endpoint: str,
        payload: dict,
        include_api_version: bool = True,
        expected_status_code: int | List[int] = 201,
        file_path: str = None,
    ) -> requests.Response:
        """
        Shortcut function to make a POST request to the specified endpoint, including
        a file and using Authentication object.

        Args:
            endpoint (str): The endpoint to call using POST HTTP verb.
            payload (dict): Payload to send with the POST request as secretmetadata.
            include_api_version (bool): Whether to include API version in request URL.
            expected_status_code (int | List[int], optional): Expected status code to
                consider the request was successful.
            file_path (str, optional): The path to the file to be uploaded. If not
                provided, no file will be uploaded. The FileName key is automatically
                included in request body using file_path.

        Returns:
            request.Response: Response object.

        Raises:
            exceptions.CreationError: Raised when creation failed, because response
                status code is different to expected_status_code.
            FileNotFoundError: If the provided file_path does not point to an existing
                file.
            requests.exceptions.RequestException: If there is an error during the file
                upload.
        """
        url = self._create_url(endpoint, include_api_version)

        utils.print_log(self._logger, f"Calling URL: {url}", logging.DEBUG)

        with open(file_path, "rb") as file:
            file_name = os.path.basename(file_path)
            payload["FileName"] = file_name

            files = {
                # The None value is used as the filename, since it's not a file upload:
                "secretmetadata": (None, json.dumps(payload), "application/json"),
                "files": (file_name, file, "application/octet-stream"),
            }

            response = self._authentication._req.post(
                url=url,
                files=files,
                timeout=(
                    self._authentication._timeout_connection_seconds,
                    self._authentication._timeout_request_seconds,
                ),
            )

            if not self.__is_successful_response(
                response.status_code, expected_status_code
            ):
                if not self._authentication.sign_app_out():
                    utils.print_log(
                        self._logger, "Error in post file request", logging.ERROR
                    )
                raise exceptions.CreationError(
                    f"Error running post file request, message: {response.text}"
                    f", statuscode: {response.status_code}"
                )

            return response

    def _run_delete_request(
        self,
        endpoint: str,
        expected_status_code: int | List[int] = 200,
    ) -> None:
        """
        Shortcut function to make a DELETE request to the specified endpoint using
        Authentication object.

        Args:
            endpoint (str): The endpoint to call using DELETE HTTP verb.
            include_api_version (bool): Whether to include API version in request URL.
            expected_status_code (int | List[int], optional): Expected status code to
                consider the request was successful.

        Returns:
            request.Response: Response object.

        Raises:
            exceptions.DeletionError: When delete request fails.
        """
        url = self._create_url(endpoint, include_api_version=False)

        utils.print_log(self._logger, f"DELETE request to URL: {url}", logging.DEBUG)

        response = self._authentication._req.delete(
            url,
            timeout=(
                self._authentication._timeout_connection_seconds,
                self._authentication._timeout_request_seconds,
            ),
        )

        if not self.__is_successful_response(
            response.status_code, expected_status_code
        ):
            if not self._authentication.sign_app_out():
                utils.print_log(
                    self._logger, "Error running delete request", logging.ERROR
                )
            raise exceptions.DeletionError(
                f"Error running delete request, message: {response.text}, statuscode: "
                f"{response.status_code}"
            )

    def _run_put_request(
        self,
        endpoint: str,
        payload: dict,
        include_api_version: bool = True,
        expected_status_code: int | List[int] = 200,
    ) -> requests.Response:
        """
        Shortcut function to make a PUT request to the specified endpoint using
        Authentication object.

        Args:
            endpoint (str): The endpoint to call using PUT HTTP verb.
            payload (dict): Payload to send with the PUT request.
            include_api_version (bool): Whether to include API version in request URL.
            expected_status_code (int | List[int], optional): Expected status code to
                consider the request was successful.

        Returns:
            request.Response: Response object.

        Raises:
            exceptions.UpdateError: Raised when update failed, because response
                status code is different to expected_status_code.
        """
        url = self._create_url(endpoint, include_api_version)

        utils.print_log(self._logger, f"Calling URL: {url}", logging.DEBUG)

        response = self._authentication._req.put(
            url=url,
            json=payload,
            timeout=(
                self._authentication._timeout_connection_seconds,
                self._authentication._timeout_request_seconds,
            ),
        )

        if not self.__is_successful_response(
            response.status_code, expected_status_code
        ):
            if not self._authentication.sign_app_out():
                utils.print_log(self._logger, "Error in put request", logging.ERROR)
            raise exceptions.UpdateError(
                f"Error running put request, message: {response.text}, statuscode: "
                f"{response.status_code}"
            )

        return response

    def _run_put_file_request(
        self,
        endpoint: str,
        payload: dict,
        include_api_version: bool = True,
        expected_status_code: int | List[int] = 204,
        file_path: str = None,
    ) -> requests.Response:
        """
        Shortcut function to make a PUT request to the specified endpoint, including
        a file and using Authentication object.

        Args:
            endpoint (str): The endpoint to call using PUT HTTP verb.
            payload (dict): Payload to send with the PUT request as secretmetadata.
            include_api_version (bool): Whether to include API version in request URL.
            expected_status_code (int | List[int], optional): Expected status code to
                consider the request was successful.
            file_path (str, optional): The path to the file to be uploaded. If not
                provided, no file will be uploaded. The FileName key is automatically
                included in request body using file_path.

        Returns:
            request.Response: Response object.

        Raises:
            exceptions.UpdateError: Raised when update failed, because response
                status code is different to expected_status_code.
            FileNotFoundError: If the provided file_path does not point to an existing
                file.
            requests.exceptions.RequestException: If there is an error during the file
                upload.
        """
        url = self._create_url(endpoint, include_api_version)

        utils.print_log(self._logger, f"Calling URL: {url}", logging.DEBUG)

        with open(file_path, "rb") as file:
            file_name = os.path.basename(file_path)
            payload["FileName"] = file_name

            files = {
                # The None value is used as the filename, since it's not a file upload:
                "secretmetadata": (None, json.dumps(payload), "application/json"),
                "files": (file_name, file, "application/octet-stream"),
            }

            response = self._authentication._req.put(
                url=url,
                files=files,
                timeout=(
                    self._authentication._timeout_connection_seconds,
                    self._authentication._timeout_request_seconds,
                ),
            )

            if not self.__is_successful_response(
                response.status_code, expected_status_code
            ):
                if not self._authentication.sign_app_out():
                    utils.print_log(
                        self._logger, "Error in put file request", logging.ERROR
                    )
                raise exceptions.UpdateError(
                    f"Error running put file request, message: {response.text}"
                    f", statuscode: {response.status_code}"
                )

            return response

    def _create_url(self, endpoint: str, include_api_version: bool = True) -> str:
        """
        Constructs a complete URL by appending the specified endpoint to the base API
        URL.
        Optionally includes the API version as a query parameter.

        Args:
            endpoint (str): The endpoint to be appended to the base API URL. This should
                start with a '/'.
            include_api_version (bool): Flag to determine whether to include the API
                version in the URL. Defaults to True. If True and the API version is
                specified in the authentication object, it will be included as a query
                parameter.

        Returns:
            str: The fully constructed URL.

        Example:
            If the base API URL is "http://api.example.com", the endpoint is "/data",
            and the API version is "v1", then the resulting URL will be
            "http://api.example.com/data?version=v1" if  include_api_version is True.
            If include_api_version is False, the resulting URL will be
            "http://api.example.com/data".
        """
        url = f"{self._authentication._api_url}{endpoint}"

        if self._authentication._api_version and include_api_version:
            params = {"version": self._authentication._api_version}
            url = self.add_api_version(url=url, new_params=params)
        return url

    def add_api_version(self, url: str, new_params: dict) -> str:
        """
        Appends or updates query parameters in the given URL.

        This function parses the provided URL, updates its query parameters with the new
        parameters provided, and reconstructs the URL.

        Args:
            url (str): The original URL to which parameters will be added or updated.
            new_params (dict): A dictionary of parameters to add or update in the URL.
                The dictionary should have parameter names as keys and parameter values
                as values.

        Returns:
            str: The new URL with updated query parameters.

        Example:
            Given the URL "http://example.com/data?filter=old" and new_params
            {"version": "v1", "filter": "new"}, the function returns
            "http://example.com/data?filter=new&version=v1".
        """
        parsed_url = urlparse(url)
        query_params = parse_qs(parsed_url.query)
        query_params.update(new_params)

        query_params = {k: v[0] if len(v) == 1 else v for k, v in query_params.items()}
        query = urlencode(query_params, doseq=True)
        new_url = urlunparse(parsed_url._replace(query=query))
        return new_url

    def make_query_string(self, params: dict) -> str:
        """
        Constructs a query string from a dictionary of parameters, excluding any
        parameters with a value of None.

        This function filters out any key-value pairs in the input dictionary where the
        value is None, and then encodes the remaining parameters into a URL-encoded
        query string.

        Args:
            params (dict): A dictionary containing the parameters to be included in the
                query string. Keys are parameter names, and values are parameter values.
                Parameters with None values are excluded from the resulting query
                string.

        Returns:
            str: A URL-encoded query string constructed from the provided parameters. If
                all parameters are None, returns an empty string.

        Example:
            Given params = {"name": "John", "age": 30, "city": None}, the function
            returns "name=John&age=30".
        """
        query_params = {k: v for k, v in params.items() if v is not None}
        query_string = urlencode(query_params)
        return query_string

    def get_request_body_version(self, body: dict, endpoint: dict, version: str = None):
        selected_version = version or self._authentication._api_version
        requests_body = body.get(endpoint, {}).get(selected_version)
        return requests_body

    def generate_request_body(self, dict_structure: dict, **kwargs) -> dict:
        """
        Transforms keyword arguments into a structured dictionary based on a given
        schema.

        This function maps keys from kwargs to a predefined dictionary structure,
        converting keys to CamelCase and filtering out any keys with None values. It
        handles nested lists of dictionaries by ensuring only specified sub-keys are
        included.

        Args:
            dict_structure (dict): A dictionary defining the desired structure and
                types of the output dictionary. The keys should be in CamelCase and the
                values should indicate the expected type, which could be a simple type
                (like str or int) or a list of dictionaries specifying sub-structure.
            **kwargs: Arbitrary keyword arguments that represent the data to be
                transformed according to `dict_structure`. Keys should be in snake_case
                or similar formats and will be converted to CamelCase.

        Returns:
            dict: A dictionary containing the structured data extracted and transformed
                from kwargs based on `dict_structure`. Only keys that exist in the
                `dict_structure` and have non-None values in `kwargs` are included.

        Example:
            >>> dict_structure = {
            >>>     'Title': str,
            >>>     'Description': str,
            >>>     'Username': str,
            >>>     'Password': str,
            >>>     'OwnerType': str,
            >>>     'Owners': [{'OwnerId': int, 'Owner': str, 'Email': str}]
            >>> }
            >>> kwargs = {
            >>>     'title': 'Example',
            >>>     'description': 'An example',
            >>>     'username': 'user1',
            >>>     'password': 'pass',
            >>>     'owner_type': 'User',
            >>>     'owners': [{'owner_id': 1, 'owner': 'Owner1'}]
            >>> }
            >>> output = generate_request_body(dict_structure, **kwargs)
            >>> print(output)
            >>> # Output: {
            >>> #     'Title': 'Example',
            >>> #     'Description': 'An example',
            >>> #     'Username': 'user1',
            >>> #     'Password': 'pass',
            >>> #     'OwnerType': 'User',
            >>> #     'Owners': [{'OwnerId': 1, 'Owner': 'Owner1'}]
            >>> # }
        """

        def simplify_key(key):
            """Converts a key to lowercase and removes underscores."""
            return key.replace("_", "").lower()

        def build_key_map(structure):
            """Builds a mapping of simplified keys to original keys."""
            return {simplify_key(key): key for key in structure.keys()}

        def transform_nested_list(nested_structure, nested_list):
            """Transforms a list of nested dictionaries."""
            simplified_nested_structure = {
                simplify_key(k): k for k in nested_structure[0].keys()
            }
            return [
                {
                    simplified_nested_structure[simplify_key(k)]: v
                    for k, v in item.items()
                    if simplify_key(k) in simplified_nested_structure
                }
                for item in nested_list
            ]

        def transform_value(value, value_type):
            """Transforms a value based on its type in the structure."""
            if isinstance(value_type, list) and isinstance(value, list):
                # Handle nested lists of dictionaries
                return transform_nested_list(value_type, value)
            return value

        # Build the key map for the top-level structure
        key_map = build_key_map(dict_structure)

        # Transform the kwargs into the output structure
        output = {}
        for kwarg_key, kwarg_value in kwargs.items():
            if kwarg_value is None:
                continue

            simplified_key = simplify_key(kwarg_key)

            if simplified_key in key_map:
                original_key = key_map[simplified_key]
                value_type = dict_structure[original_key]
                output[original_key] = transform_value(kwarg_value, value_type)

        return output



"""Utils module"""

import logging

from cerberus import Validator

from secrets_safe_library import exceptions


def print_log(logger, log_message, level):
    """
    Print a log message
    Arguments:
        Logger
        Log level
        Log message
    Returns:
    """
    if logger:
        if level == logging.DEBUG:
            logger.debug(log_message)
        elif level == logging.INFO:
            logger.info(log_message)
        elif level == logging.ERROR:
            logger.error(log_message)
        elif level == logging.WARN:
            logger.warning(log_message)


def validate_inputs(inputs):
    """
    Validate inputs
    Arguments:
        inputs (dict)
    Returns:
        inputs (dict)
    """

    schema = {
        "api_url": {
            "type": "string",
            "regex": r"^(https?|http)://[^\s/$.?#].[^\s]*$",
            "required": True,
        }
    }

    if "api_version" in inputs:
        schema.update(
            {
                "api_version": {
                    "type": "string",
                    "minlength": 3,
                    "maxlength": 3,
                    "required": False,
                }
            }
        )

    if "api_key" in inputs:
        schema.update(
            {
                "api_key": {
                    "type": "string",
                    "minlength": 128,
                    "maxlength": 263,
                    "required": True,
                }
            }
        )
    else:
        schema.update(
            {
                "client_id": {
                    "type": "string",
                    "minlength": 36,
                    "maxlength": 40,
                    "required": True,
                },
                "client_secret": {
                    "type": "string",
                    "minlength": 36,
                    "maxlength": 64,
                    "required": True,
                },
            }
        )

    v = Validator(schema)

    if v.validate(inputs):
        if "/BeyondTrust/api/public/v" not in inputs["api_url"]:
            raise exceptions.OptionsError(
                "invalid API URL, it must contains /BeyondTrust/api/public/v as part of"
                " the path"
            )

        inputs["api_url"] = inputs["api_url"].strip()

    else:
        raise exceptions.OptionsError(f"Errors: {v.errors}")

    return inputs


def prepare_certificate_info(certificate, certificate_key):
    """
    Validate certificate and certificate key
    Arguments:
        certificate
        certificate_key
    Returns:
        certificate
        certificate_key

    """

    if certificate and certificate_key:

        certificate_length_in_bits = len(certificate.encode("utf-8")) * 8

        if certificate_length_in_bits > 32768:
            message = "Invalid length for certificate, the maximum size is 32768 bits"
            raise exceptions.OptionsError(message)

        certificate_key_length_in_bits = len(certificate_key.encode("utf-8")) * 8

        if certificate_key_length_in_bits > 32768:
            message = (
                "Invalid length for certificate key, the maximum size is 32768 bits"
            )
            raise exceptions.OptionsError(message)

        if "BEGIN CERTIFICATE" not in certificate:
            raise exceptions.OptionsError("Bad certificate content")

        if "BEGIN PRIVATE KEY" not in certificate_key:
            raise exceptions.OptionsError("Bad certificate key content")

        certificate = certificate.replace(r"\n", "\n")
        certificate_key = certificate_key.replace(r"\n", "\n")

        certificate = f"{certificate}\n"
        certificate_key = f"{certificate_key}\n"

        return certificate, certificate_key

    return "", ""



"""Assets module, all the logic to manage assets from PS API"""

import ipaddress
import logging
from typing import List

from cerberus import Validator

from secrets_safe_library import exceptions, utils
from secrets_safe_library.authentication import Authentication
from secrets_safe_library.core import APIObject


class CustomValidator(Validator):
    def _validate_is_ip(self, is_ip, field, value):
        """Custom validation rule to check if a value is a valid IP address.

        The rule's arguments are validated against this schema:
        {'type': 'boolean'}
        """
        if is_ip and value:
            try:
                ipaddress.ip_address(value)  # Validate IP address
            except ValueError:
                self._error(field, f"'{value}' is not a valid IP address.")


class Asset(APIObject):

    def __init__(self, authentication: Authentication, logger: logging.Logger = None):
        super().__init__(authentication, logger)

        # Schema rules used for validations
        self._schema = {
            "workgroup_id": {"type": "integer", "nullable": True},
            "workgroup_name": {"type": "string", "nullable": True},
            "ip_address": {
                "type": "string",
                "is_ip": True,
                "maxlength": 45,
                "nullable": False,
                "required": True,
            },
            "asset_name": {
                "type": "string",
                "maxlength": 128,
                "nullable": True,
                "required": False,
            },
            "dns_name": {
                "type": "string",
                "maxlength": 255,
                "nullable": True,
                "required": False,
            },
            "domain_name": {
                "type": "string",
                "maxlength": 64,
                "nullable": True,
                "required": False,
            },
            "mac_address": {
                "type": "string",
                "maxlength": 128,
                "nullable": True,
                "required": False,
            },
            "asset_type": {
                "type": "string",
                "maxlength": 64,
                "nullable": True,
                "required": False,
            },
            "description": {
                "type": "string",
                "maxlength": 255,
                "nullable": True,
                "required": False,
            },
            "operating_system": {
                "type": "string",
                "maxlength": 255,
                "nullable": True,
                "required": False,
            },
        }
        self._validator = CustomValidator(self._schema)

    def list_assets(
        self,
        workgroup_id: int = None,
        workgroup_name: str = None,
        limit: int = None,
        offset: int = None,
    ) -> list:
        """
        Returns a list of assets matching specified Workgroup ID or Name.
        parameters.

        API:
            - GET Workgroups/{workgroupID}/Assets
            - GET Workgroups/{workgroupName}/Assets

        Args:
            workgroup_id (int, optional): The Workgroup ID, if want to search by
                Workgroup ID.
            workgroup_name (str, optional): The Workgroup name, if want to search by
                Workgroup name.
            limit (int, optional): limit the results.
            offset (int, optional): skip the first (offset) number of assets.

        Returns:
            list: List of assets matching specified Workgroup ID or Name.
        """

        params = {
            "limit": limit,
            "offset": offset,
        }
        query_string = self.make_query_string(params)

        if workgroup_id:
            endpoint = f"/workgroups/{workgroup_id}/assets?{query_string}"
        elif workgroup_name:
            endpoint = f"/workgroups/{workgroup_name}/assets?{query_string}"
        else:
            raise exceptions.OptionsError(
                "Either workgroup_id or workgroup_name is required"
            )

        utils.print_log(
            self._logger, f"Calling list_assets endpoint: {endpoint}", logging.DEBUG
        )
        response = self._run_get_request(endpoint)

        return response.json()

    def get_asset_by_id(self, asset_id: str) -> dict:
        """
        Returns an asset by ID.

        API: GET Assets/{id}

        Args:
            asset_id (str): The asset ID (GUID).

        Returns:
            dict: Asset object.
        """

        endpoint = f"/assets/{asset_id}"

        utils.print_log(
            self._logger,
            f"Calling get_asset_by_id endpoint: {endpoint}",
            logging.DEBUG,
        )
        response = self._run_get_request(endpoint, include_api_version=False)

        return response.json()

    def get_asset_by_workgroup_name(self, workgroup_name: str, asset_name: str) -> dict:
        """
        Returns an asset by Workgroup name and asset name.

        API: GET Workgroups/{workgroupName}/Assets?name={name}

        Args:
            workgroup_name (str): Name of the Workgroup.
            asset_name (str): Name of the asset.

        Returns:
            dict: Asset object.
        """

        params = {"name": asset_name}
        query_string = self.make_query_string(params)

        endpoint = f"/workgroups/{workgroup_name}/assets?{query_string}"

        utils.print_log(
            self._logger,
            f"Calling get_asset_by_workgroup_name endpoint: {endpoint}",
            logging.DEBUG,
        )
        response = self._run_get_request(endpoint, include_api_version=False)

        return response.json()

    def list_asset_attributes(self, asset_id: int) -> List[dict]:
        """
        Returns a list of attributes for a specific asset.

        API: GET Assets/{assetID}/Attributes

        Args:
            asset_id (int): The asset ID.

        Returns:
            List[dict]: List of attributes for the specified asset.
        """

        endpoint = f"/assets/{asset_id}/attributes"

        utils.print_log(
            self._logger,
            f"Calling list_asset_attributes endpoint: {endpoint}",
            logging.DEBUG,
        )
        response = self._run_get_request(endpoint, include_api_version=False)

        return response.json()

    def search_assets(
        self,
        asset_name: str = None,
        dns_name: str = None,
        domain_name: str = None,
        ip_address: str = None,
        mac_address: str = None,
        asset_type: str = None,
        limit: int = 100000,
        offset: int = 0,
    ) -> list:
        """
        Returns a list of assets that match the given search criteria.

        At least one request body property should be provided; any property not
        provided is ignored. All search criteria is case insensitive and is an exact
        match (equality), except for IPAddress.

        API: POST Assets/Search

        Args:
            asset_name (str, optional): The Asset name, if you want to search by
                Asset name.
            dns_name (str, optional): The DNS name, if you want to search by DNS name.
            domain_name (str, optional): The Domain name, if you want to search by
                Domain name.
            ip_address (str, optional): The IP address, if you want to search by IP
                address.
            mac_address (str, optional): The MAC address, if you want to search by MAC
                address.
            asset_type (str, optional): The Asset type, if you want to search by Asset
                type.
            limit (int, optional): Limit the number of results returned.
            offset (int, optional): Skip the first (offset) number of assets in the
                results.

        Returns:
            list: List of assets matching specified parameters.
        """

        if not any(
            [asset_name, dns_name, domain_name, ip_address, mac_address, asset_type]
        ):
            raise exceptions.OptionsError(
                "At least one of the following fields must be provided: "
                "asset_name, dns_name, domain_name, ip_address, mac_address, asset_type"
            )

        params = {
            "limit": limit,
            "offset": offset,
        }
        query_string = self.make_query_string(params)

        endpoint = f"/assets/search?{query_string}"

        utils.print_log(
            self._logger, f"Calling list_assets endpoint: {endpoint}", logging.DEBUG
        )

        body = {
            "AssetName": asset_name,
            "DnsName": dns_name,
            "DomainName": domain_name,
            "IPAddress": ip_address,
            "MacAddress": mac_address,
            "AssetType": asset_type,
        }

        req_body = {key: value for key, value in body.items() if value is not None}
        response = self._run_post_request(endpoint, req_body, expected_status_code=200)

        return response.json()

    def create_asset(
        self,
        *,
        workgroup_id: int = None,
        workgroup_name: str = None,
        ip_address: str = None,
        asset_name: str = None,
        dns_name: str = None,
        domain_name: str = None,
        mac_address: str = None,
        asset_type: str = None,
        description: str = None,
        operating_system: str = None,
    ) -> dict:
        """
        Creates a new asset in the Workgroup, referenced by ID (workgroup_id) or
        Workgroup Name.

        If both workgroup_id and workgroup_name are provided, the workgroup_id is used.

        API:
            - POST Workgroups/{workgroupID}/Assets
            - POST Workgroups/{workgroupName}/Assets

        Args:
            workgroup_id (int): Workgroup ID.
            ip_address (str): Asset IP address. Required. Max string length is 45.
            asset_name (str, optional): Asset name. If not provided, a padded IP address
                is used. Max string length is 128.
            dns_name (str, optional): Asset DNS name. Max string length is 255.
            domain_name (str, optional): Asset domain name. Max string length is 64.
            mac_address (str, optional): Asset MAC address. Max string length is 128.
            asset_type (str, optional): Asset type. Max string length is 64.
            description (str, optional): Asset description. Only updated if the version
                in the URL is 3.1 or greater. Max string length is 255.
            operating_system (str, optional): Asset operating system. Max string length
                is 255.

        Returns:
            dict: Asset object.
        """

        attributes = {
            "workgroup_id": workgroup_id,
            "ip_address": ip_address,
            "asset_name": asset_name,
            "dns_name": dns_name,
            "domain_name": domain_name,
            "mac_address": mac_address,
            "asset_type": asset_type,
            "description": description,
            "operating_system": operating_system,
        }

        if not self._validator.validate(attributes, update=True):
            raise exceptions.OptionsError(f"Please check: {self._validator.errors}")

        if workgroup_id is None and workgroup_name is None:
            raise exceptions.OptionsError(
                "Either workgroup_id or workgroup_name is required"
            )

        identifier = workgroup_id if workgroup_id is not None else workgroup_name
        endpoint = f"/workgroups/{identifier}/assets"

        utils.print_log(
            self._logger, f"Calling create_asset endpoint: {endpoint}", logging.DEBUG
        )

        body = {
            "IPAddress": ip_address,
            "AssetName": asset_name,
            "DnsName": dns_name,
            "DomainName": domain_name,
            "MacAddress": mac_address,
            "AssetType": asset_type,
            "Description": description,
            "OperatingSystem": operating_system,
        }

        req_body = {key: value for key, value in body.items() if value is not None}
        response = self._run_post_request(endpoint, req_body, expected_status_code=201)

        return response.json()

    def update_asset(
        self,
        asset_id: int,
        *,
        workgroup_id: int = None,
        asset_name: str = None,
        dns_name: str = None,
        domain_name: str = None,
        ip_address: str = None,
        mac_address: str = None,
        asset_type: str = None,
        description: str = None,
        operating_system: str = None,
    ) -> dict:
        """
        Updates an existing asset by ID.

        If both workgroup_id and workgroup_name are provided, the workgroup_id is used.

        API:
            - PUT Assets/{id}

        Args:
            asset_id (int): ID of the asset.
            workgroup_id (int): Workgroup ID.
            ip_address (str): Asset IP address. Required. Max string length is 45.
            asset_name (str, optional): Asset name. If not provided, a padded IP address
                is used. Max string length is 128.
            dns_name (str, optional): Asset DNS name. Max string length is 255.
            domain_name (str, optional): Asset domain name. Max string length is 64.
            mac_address (str, optional): Asset MAC address. Max string length is 128.
            asset_type (str, optional): Asset type. Max string length is 64.
            description (str, optional): Asset description. Only updated if the version
                in the URL is 3.1 or greater. Max string length is 255.
            operating_system (str, optional): Asset operating system. Max string length
                is 255.

        Returns:
            dict: Asset object.
        """

        attributes = {
            "workgroup_id": workgroup_id,
            "ip_address": ip_address,
            "asset_name": asset_name,
            "dns_name": dns_name,
            "domain_name": domain_name,
            "mac_address": mac_address,
            "asset_type": asset_type,
            "description": description,
            "operating_system": operating_system,
        }

        if not self._validator.validate(attributes, update=True):
            raise exceptions.OptionsError(f"Please check: {self._validator.errors}")

        if not workgroup_id:
            raise exceptions.OptionsError("workgroup_id is required")

        endpoint = f"/assets/{asset_id}"

        utils.print_log(
            self._logger, f"Calling update_asset endpoint: {endpoint}", logging.DEBUG
        )

        body = {
            "WorkgroupID": workgroup_id,
            "IPAddress": ip_address,
            "AssetName": asset_name,
            "DnsName": dns_name,
            "DomainName": domain_name,
            "MacAddress": mac_address,
            "AssetType": asset_type,
            "Description": description,
            "OperatingSystem": operating_system,
        }

        req_body = {key: value for key, value in body.items() if value is not None}
        response = self._run_put_request(endpoint, req_body, expected_status_code=200)

        return response.json()

    def delete_asset_by_id(self, asset_id: int) -> None:
        """
        Deletes an asset by ID.

        API: DELETE Assets/{id}

        Args:
            asset_id (int): The asset ID.

        Returns:
            None
        """

        endpoint = f"/assets/{asset_id}"

        utils.print_log(
            self._logger,
            f"Calling delete_asset_by_id endpoint: {endpoint}",
            logging.DEBUG,
        )
        self._run_delete_request(endpoint, expected_status_code=200)

    def assign_asset_attribute(self, asset_id: int, attribute_id: int) -> dict:
        """
        Assigns an attribute to an asset.

        API: POST Assets/{assetID}/Attributes/{attributeID}

        Args:
            asset_id (int): The asset ID.
            attribute_id (int): The attribute ID.

        Returns:
            dict: Asset object.
        """

        endpoint = f"/assets/{asset_id}/attributes/{attribute_id}"

        utils.print_log(
            self._logger,
            f"Calling assign_asset_attribute endpoint: {endpoint}",
            logging.DEBUG,
        )

        response = self._run_post_request(
            endpoint, payload={}, expected_status_code=201
        )

        return response.json()

    def delete_asset_attribute(self, asset_id: int, attribute_id: int) -> None:
        """
        Deletes an attribute from an asset.

        API: DELETE Assets/{assetID}/Attributes/{attributeID}

        Args:
            asset_id (int): The asset ID.
            attribute_id (int): The attribute ID.

        Returns:
            None
        """

        endpoint = f"/assets/{asset_id}/attributes/{attribute_id}"

        utils.print_log(
            self._logger,
            f"Calling delete_asset_attribute endpoint: {endpoint}",
            logging.DEBUG,
        )
        self._run_delete_request(endpoint, expected_status_code=200)



"""
This file contains the fields mapping for each endpoint and version related to Safes.

More info for the fields included in each endpoint and version:
- https://docs.beyondtrust.com/bips/docs/passwordsafe-api
- https://docs.beyondtrust.com/bips/docs/secrets-safe-api
"""

from secrets_safe_library.constants.endpoints import (
    GET_MANAGED_ACCOUNTS_ID,
    GET_MANAGED_SYSTEMS_SYSTEM_ID_MANAGED_ACCOUNTS,
)
from secrets_safe_library.constants.versions import Version

fields = {
    GET_MANAGED_SYSTEMS_SYSTEM_ID_MANAGED_ACCOUNTS: {
        Version.DEFAULT: [
            "ManagedAccountID",
            "ManagedSystemID",
            "DomainName",
            "AccountName",
            "DistinguishedName",
            "PasswordFallbackFlag",
            "UserPrincipalName",
            "SAMAccountName",
            "LoginAccountFlag",
            "Description",
            "PasswordRuleID",
            "ApiEnabled",
            "ReleaseNotificationEmail",
            "ChangeServicesFlag",
            "RestartServicesFlag",
            "ChangeTasksFlag",
            "ReleaseDuration",
            "MaxReleaseDuration",
            "ISAReleaseDuration",
            "MaxConcurrentRequests",
            "AutoManagementFlag",
            "DSSAutoManagementFlag",
            "CheckPasswordFlag",
            "ResetPasswordOnMismatchFlag",
            "ChangePasswordAfterAnyReleaseFlag",
            "ChangeFrequencyType",
            "ChangeFrequencyDays",
            "ChangeTime",
            "ParentAccountID",
            "IsSubscribedAccount",
            "LastChangeDate",
            "NextChangeDate",
            "IsChanging",
            "ChangeState",
            "UseOwnCredentials",
            "WorkgroupID",
            "ChangeIISAppPoolFlag",
            "RestartIISAppPoolFlag",
            "ChangeWindowsAutoLogonFlag",
            "ChangeComPlusFlag",
            "ChangeDComFlag",
            "ChangeSComFlag",
        ]
    },
    GET_MANAGED_ACCOUNTS_ID: {
        Version.DEFAULT: [
            "ManagedAccountID",
            "ManagedSystemID",
            "DomainName",
            "AccountName",
            "DistinguishedName",
            "PasswordFallbackFlag",
            "UserPrincipalName",
            "SAMAccountName",
            "LoginAccountFlag",
            "Description",
            "PasswordRuleID",
            "ApiEnabled",
            "ReleaseNotificationEmail",
            "ChangeServicesFlag",
            "RestartServicesFlag",
            "ChangeTasksFlag",
            "ReleaseDuration",
            "MaxReleaseDuration",
            "ISAReleaseDuration",
            "MaxConcurrentRequests",
            "AutoManagementFlag",
            "DSSAutoManagementFlag",
            "CheckPasswordFlag",
            "ResetPasswordOnMismatchFlag",
            "ChangePasswordAfterAnyReleaseFlag",
            "ChangeFrequencyType",
            "ChangeFrequencyDays",
            "ChangeTime",
            "ParentAccountID",
            "IsSubscribedAccount",
            "LastChangeDate",
            "NextChangeDate",
            "IsChanging",
            "ChangeState",
            "UseOwnCredentials",
            "WorkgroupID",
            "ChangeIISAppPoolFlag",
            "RestartIISAppPoolFlag",
            "ObjectID",
        ]
    },
}



"""Folders module, all the logic to manage folders from PS API"""

import logging

from cerberus import Validator

from secrets_safe_library import exceptions, utils
from secrets_safe_library.authentication import Authentication
from secrets_safe_library.core import APIObject


class Folder(APIObject):

    def __init__(self, authentication: Authentication, logger: logging.Logger = None):
        super().__init__(authentication, logger)

        # Schema rules used for validations
        self._schema = {
            "description": {"type": "string", "maxlength": 256, "nullable": True},
        }
        self._validator = Validator(self._schema)

    def get_folder_by_id(self, folder_id: str) -> dict:
        """
        Find a folder by ID.

        API: GET Secrets-Safe/Folders/{id}

        Args:
            folder_id (str): The folder ID (GUID).

        Returns:
            dict: Folder object according requested API version.
        """

        endpoint = f"/secrets-safe/folders/{folder_id}"

        utils.print_log(
            self._logger,
            f"Calling get_folder_by_id endpoint: {endpoint}",
            logging.DEBUG,
        )
        response = self._run_get_request(endpoint, include_api_version=False)

        return response.json()

    def list_folders(
        self,
        folder_name: str = None,
        folder_path: str = None,
        include_subfolders: bool = True,
        root_only: bool = False,
        folder_owner_id: str = None,
        limit: int = None,
        offset: int = None,
    ) -> list:
        """
        Returns a list of folders with the option to filter the list using query
        parameters.

        API: GET Secrets-Safe/Folders/

        Args:
            folder_name (str, optional): The partial name of the folder. Defaults to
            None.
            folder_path (str, optional): Child folders are also included. Separator is
            '/'. Defaults to None.
            include_subfolders (bool, optional): Indicate whether to include the
            subfolder. Defaults to True.
            root_only (bool, optional): The results only include those folders at the
            root level. Defaults to False.
            folder_owner_id (str, optional): Filter results by the folders which are
            owned by the given FolderOwnerId. Defaults to None.
            limit (int, optional): limit the results.
            offset (int, optional): skip the first (offset) number of folders.

        Returns:
            list: List of folders matching specified arguments.
        """

        params = {
            "folderName": folder_name,
            "folderPath": folder_path,
            "includeSubfolders": include_subfolders,
            "rootOnly": root_only,
            "folderOwnerId": folder_owner_id,
            "limit": limit,
            "offset": offset,
        }
        query_string = self.make_query_string(params)
        endpoint = f"/secrets-safe/folders?{query_string}"

        utils.print_log(
            self._logger, f"Calling list_folders endpoint: {endpoint}", logging.DEBUG
        )
        response = self._run_get_request(endpoint)

        return response.json()

    def create_folder(
        self,
        name: str,
        parent_id: str,
        description: str = "",
        user_group_id: int = None,
    ) -> dict:
        """
        Creates a new Secrets Safe folder for the given user group.

        API: POST Secrets-Safe/Folders/

        Args:
            name (str): The folder name.
            parent_id (str): The parent folder/safe ID (GUID).
            description (str, optional): The folder description.
            user_group_id (int, optional): The user group ID.

        Returns:
            dict: Created Folder object.
        """

        attributes = {"description": description}

        if not self._validator.validate(attributes, update=True):
            raise exceptions.OptionsError(f"Please check: {self._validator.errors}")

        payload = {
            "Name": name,
            "Description": description,
            "ParentId": parent_id,
            "UserGroupId": user_group_id,
        }
        endpoint = "/secrets-safe/folders"

        utils.print_log(
            self._logger, f"Calling create folder endpoint: {endpoint}", logging.DEBUG
        )
        response = self._run_post_request(endpoint, payload, include_api_version=False)

        return response.json()

    def delete_folder_by_id(self, folder_id: str) -> None:
        """
        Delete a folder by ID.

        API: DELETE Secrets-Safe/Folders/{id}

        Args:
            folder_id (str): The folder ID (GUID).

        Returns:
            None: If deletion is successful no exceptions.DeletionError is raised.
        """

        endpoint = f"/secrets-safe/folders/{folder_id}"

        utils.print_log(
            self._logger,
            f"Calling delete_folder_by_id endpoint: {endpoint}",
            logging.DEBUG,
        )
        self._run_delete_request(endpoint)



"""Custom Exceptions"""


class AuthenticationFailure(Exception):
    """Exception raised for errors in Autentication

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message="AuthenticationFailure error"):
        self.message = message
        super().__init__(self.message)


class OptionsError(Exception):
    """Exception raised for errors in Options

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message="OptionsError error"):
        self.message = message
        super().__init__(self.message)


class LookupError(Exception):
    """Exception raised for errors looking secrets

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message="LookupError error"):
        self.message = message
        super().__init__(self.message)


class CreationError(Exception):
    """Exception raised when object was not created in API

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message="CreationError error"):
        self.message = message
        super().__init__(self.message)


class UpdateError(Exception):
    """Exception raised when object was not updated in API

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message="UpdateError error"):
        self.message = message
        super().__init__(self.message)


class DeletionError(Exception):
    """Exception raised when object was not deleted in API

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message="DeletionError error"):
        self.message = message
        super().__init__(self.message)


class IncompleteArgumentsError(Exception):
    """Exception raised when expected function arguments are missing.

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message="IncompleteArgumentsError error"):
        self.message = message
        super().__init__(self.message)



"""
Address groups module, all the logic to manage address groups from BeyondInsight API
"""

import logging

from cerberus import Validator

from secrets_safe_library import utils
from secrets_safe_library.authentication import Authentication
from secrets_safe_library.core import APIObject


class AddressGroup(APIObject):

    def __init__(self, authentication: Authentication, logger: logging.Logger = None):
        super().__init__(authentication, logger)

        # Schema rules used for validations
        self._schema = {
            "name": {"type": "string", "maxlength": 256, "nullable": True},
            "description": {"type": "string", "maxlength": 256, "nullable": True},
        }
        self._validator = Validator(self._schema)

    def get_address_group_by_id(self, address_group_id: int) -> dict:
        """
        Find an address group by ID.

        API: GET AddressGroups/{id}

        Args:
            address_group_id (int): The address group ID.

        Returns:
            dict: Address group object according requested API version.
        """

        endpoint = f"/addressgroups/{address_group_id}"

        utils.print_log(
            self._logger,
            f"Calling get_address_group_by_id endpoint: {endpoint}",
            logging.DEBUG,
        )
        response = self._run_get_request(endpoint, include_api_version=False)

        return response.json()

    def list_address_groups(self) -> list:
        """
        List the address groups.

        API: GET AddressGroups

        Returns:
            list: List of address groups.
        """

        endpoint = "/addressgroups"

        utils.print_log(
            self._logger,
            f"Calling list_address_groups endpoint: {endpoint}",
            logging.DEBUG,
        )
        response = self._run_get_request(endpoint)

        return response.json()

    def get_address_group_by_name(self, address_group_name: str) -> dict:
        """
        Returns an address group by name.

        API: GET AddressGroups/?name={name}

        Args:
            address_group_name (str): Name of the address group.

        Returns:
            dict: Address group.
        """

        endpoint = f"/AddressGroups?name={address_group_name}"

        utils.print_log(
            self._logger,
            f"Calling get_address_group_by_name endpoint: {endpoint}",
            logging.DEBUG,
        )
        response = self._run_get_request(endpoint, include_api_version=False)

        return response.json()



"""
Endpoints list used in Integration Library (verb:endpoint)
"""

# Secrets
GET_SECRETS_SAFE_SECRETS = "get:secrets-safe/secrets"
GET_SECRETS_SAFE_SECRETS_SECRETID = "get:secrets-safe/secrets/{secretid}"
GET_SECRETS_SAFE_SECRETS_SECRETID_TEXT = "get:secrets-safe/secrets/{secretid}/text"
GET_SECRETS_SAFE_SECRETS_SECRETID_FILE = "get:secrets-safe/secrets/{secretid}/file"
GET_SECRETS_SAFE_FOLDERS_FOLDERID_SECRETS = (
    "get:secrets-safe/folders/{folderid}/secrets"
)
GET_SECRETS_SAFE_FOLDERS_FOLDERID_TEXT = "get:secrets-safe/folders/{folderid}/text"
GET_SECRETS_SAFE_FOLDERS_FOLDERID_FILE = "get:secrets-safe/folders/{folderid}/file"
POST_SECRETS_SAFE_FOLDERS_FOLDERID = "post:secrets-safe/folders/{folderid}/secrets"
POST_SECRETS_SAFE_FOLDERS_FOLDERID_TEXT = (
    "post:secrets-safe/folders/{folderid}/secrets/text"
)
POST_SECRETS_SAFE_FOLDERS_FOLDERID_FILE = (
    "post:secrets-safe/folders/{folderid}/secrets/file"
)
PUT_SECRETS_SAFE_SECRETS_SECRETID = "put:secrets-safe/secrets/{secretid}"
PUT_SECRETS_SAFE_SECRETS_SECRETID_TEXT = "put:secrets-safe/secrets/{secretid}/text"
PUT_SECRETS_SAFE_SECRETS_SECRETID_FILE = "put:secrets-safe/secrets/{secretid}/file"

# Folders
GET_SECRETS_SAFE_FOLDERS = "get:secrets-safe/folders"
GET_SECRETS_SAFE_FOLDERS_FOLDERID = "get:secrets-safe/folders/{folderid}"

# Managed Systems
GET_MANAGED_SYSTEMS = "get:managedsystems"
GET_MANAGED_SYSTEMS_MANAGEDSYSTEMID = "get:managedsystems/{managedsystemid}"
GET_MANAGED_SYSTEMS_ASSETID = "get:assets/{assetid}/managedsystems"
GET_MANAGED_SYSTEMS_DATABASEID = "get:databases/{databaseid}/managedsystems"
GET_MANAGED_SYSTEMS_FUNCTIONALACCOUNTID = (
    "get:functionalaccounts/{functionalaccountid}/managedsystems"
)
GET_MANAGED_SYSTEMS_WORKGROUPID = "get:workgroups/{workgroupid}/managedsystems"
POST_MANAGED_SYSTEMS_ASSETID = "post:assets/{assetid}/managedsystems"
POST_MANAGED_SYSTEMS_DATABASEID = "post:databases/{databaseid}/managedsystems"
POST_MANAGED_SYSTEMS_WORKGROUPID = "post:workgroups/{workgroupid}/managedsystems"
DELETE_MANAGED_SYSTEMS_MANAGEDSYSTEMID = "delete:managedsystems/{managedsystemid}"
PUT_MANAGED_SYSTEMS_MANAGEDSYSTEMID = "put:managedsystems/{managedsystemid}"

# Safes
GET_SECRETS_SAFE_SAFES = "get:secrets-safe/safes"
GET_SECRETS_SAFE_SAFES_ID = "get:secrets-safe/safes/{ID}"

# Assets
GET_WORKGROUPS_ID_ASSETS = "get:workgroups/{workgroupid}/assets"
GET_ASSETS_ID_ATTRIBUTES = "get:assets/{assetid}/attributes"
GET_ASSETS_ID = "get:assets/{id}"
GET_WORKGROUPS_NAME_ASSETS_NAME = "get:workgroups/{workgroupname}/assets?name"
POST_WORKGROUPS_WORKGROUPID_ASSETS = "post:workgroups/{workgroupid}/assets"

# Smart Rules
GET_SMARTRULES_ID = "get:smartrules/{id}"

# Organizations
GET_ORGANIZATIONS = "get:organizations"
GET_ORGANIZATIONS_ID = "get:organizations/{id}"
GET_ORGANIZATIONS_NAME = "get:organizations/{name}"

# Workgroups
GET_WORKGROUPS_ID = "get:workgroups/{id}"
GET_WORKGROUPS_NAME = "get:workgroups?name={name}"
GET_WORKGROUPS = "get:workgroups"
POST_WORKGROUPS = "post:workgroups"

# Databases
GET_DATABASES = "get:databases"
GET_DATABASES_ID = "get:databases/{id}"
GET_DATABASES_ASSET_ID = "get:assets/{assetid}/databases"
POST_DATABASES_ASSET_ID = "post:assets/{assetid}/databases"
PUT_DATABASES_ID = "put:databases/{id}"
DELETE_DATABASES_ID = "delete:databases/{id}"

# Address Groups
GET_ADDRESS_GROUPS = "get:addressgroups"
GET_ADDRESS_GROUPS_ID = "get:addressgroups/{id}"
GET_ADDRESS_GROUPS_NAME = "get:addressgroups/{name}"

# Managed Accounts
GET_MANAGED_SYSTEMS_SYSTEM_ID_MANAGED_ACCOUNTS = (
    "get:managedsystems/{systemid}/managedaccounts"
)
GET_MANAGED_ACCOUNTS_ID = "get:managedaccounts/{id}"
